from abc import ABC, abstractmethod
from typing import Optional, Union

import pandas as pd


class BaseDetector(ABC):
    """ """

    def __init__(
        self,
        threshold: float = 3.0,
    ) -> None:
        super().__init__()
        self._threshold: float = threshold
        self.anomaly_score: Optional[Union[pd.DataFrame, pd.Series]] = None

    def fit_predict(
        self, data: Union[pd.DataFrame, pd.Series]
    ) -> Union[pd.DataFrame, pd.Series]:
        """ """
        self.fit(data)
        if self.anomaly_score is not None:
            return self.anomaly_score > self._threshold
        else:
            raise ValueError('The algorithm failed to produce anomaly scores.')

    @abstractmethod
    def fit(self, data: Union[pd.DataFrame, pd.Series]) -> "BaseDetector":
        return self



import numpy as np
from numpy.lib.stride_tricks import as_strided


def rolling_window(
    X: np.ndarray,
    n_series: int,
    n_timestamps: int,
    window_size: int,
    rolling_step: int = 1,
) -> np.ndarray:
    """ """
    overlap = window_size - rolling_step
    shape_new = (
        n_series,
        (n_timestamps - overlap) // rolling_step,
        window_size // 1,
    )
    s0, s1 = X.strides
    strides_new = (s0, rolling_step * s1, s1)
    return as_strided(X, shape=shape_new, strides=strides_new)



import numpy as np

from .utils import rolling_window


class SingularSpectrumAnalysis:
    """Singular Spectrum Analysis (SSA).

    A decomposition method that splits time series into several
    components. Each of them has meaningful interpretation,
    including noise or residuals. SSA relates to the spectrum of
    eigenvalues in singular value decomposition of a covariance matrix.

    Parameters
    ----------
    window_size : int or float (default=10)
        Size of the sliding window. If float, it represents
        the percentage of the size of each time series and must be
        between 0 and 1. The window size will be computed as
        ``max(2, ceil(window_size * n_timestamps))``.

    References
    ----------
    .. [1] N. Golyandina, and A. Zhigljavsky.
            "Singular Spectrum Analysis for Time Series."
            Springer-Verlag Berlin Heidelberg (2013).
    .. [2] Donoho, David L., and Matan Gavish.
            "The optimal hard threshold for singular values is 4/√ 3."
            arXiv preprint arXiv:1305.5870 4 (2013).
    .. [3] Johann Faouzi and Hicham Janati.
            "pyts: A python package for time series classification."
            Journal of Machine Learning Research, 21(46):1−6, 2020.
    """

    def __init__(
        self,
        window_size: int,
    ) -> None:
        self._window_size = window_size

    def transform(self, X: np.ndarray) -> np.ndarray:
        """
        Transform data using SSA to extract noise or residuals.

        Parameters
        ----------
        data : pandas Series or DataFrame, shape = (n_timestamps, n_series)
            Input data.

        Returns
        -------
        X_noise : np.ndarray, shape = (n_timestamp, n_series)
            Extracted noise.
        """
        # TODO: Set batch computation
        n_series, n_timestamps = X.shape
        n_windows = int(n_timestamps - self._window_size + 1)
        # 1. Constructing trajectory matrix
        X_window = np.transpose(
            rolling_window(X, n_series, n_timestamps, self._window_size),
            axes=(0, 2, 1),
        ).copy()
        X_tranpose = np.matmul(X_window, np.transpose(X_window, axes=(0, 2, 1)))
        # 2. Decomposing trajectory covariance matrix
        w, v = np.linalg.eigh(X_tranpose)
        w, v = w[:, ::-1], v[:, :, ::-1]
        # 3. Constructing element matrices for potential signal only
        r = self._rank_calculation(w)
        X_elem = self._outer_dot(v, X_window, n_series, n_windows, r)
        if self._window_size >= n_windows:
            X_elem = np.transpose(X_elem, axes=(0, 2, 1))
            gap = self._window_size
        else:
            gap = n_windows
        # 4. Reconstructing time series (signal & noise)
        X_signal = self._diagonal_averaging(
            X_elem, n_series, n_timestamps, n_windows, gap
        )
        X_noise = X - X_signal
        return X_noise.T

    def _rank_calculation(self, w: np.ndarray, threshold: float = 2.858) -> np.ndarray:
        """
        Define the trajectory matrix ranks as a threshold
            for splitting signal and noise.

        Parameters
        ----------
        w : np.ndarray
            The eigenvalues.
        threshold : float
            The hard limit to determine the rank of trajectory matrices
            (default=2.858).

        Returns
        -------
        ranks : np.ndarray
            The rank of trajectory matrices.
        """
        rank = (w > 0).sum(axis=1)
        ranks = np.array(
            [
                (w[i, :r] >= np.median(w[i, :r]) * threshold).sum()
                for i, r in enumerate(rank)
            ]
        )
        return ranks

    def _outer_dot(
        self,
        v: np.ndarray,
        X: np.ndarray,
        n_series: int,
        n_windows: int,
        r: np.ndarray,
    ) -> np.ndarray:
        """
        Perform outer dot to construct element matrices
            for potential signal only.

        Parameters
        ----------
        v : np.ndarray
            The eigenvectors.
        X : np.ndarray
            The trajectory matrices.
        n_series : int
            Number of time series (metrics).
        n_windows : int
            The number of rows of the trajectory matrix.
        r : np.ndarray
            The rank of trajectory matrices.

        Returns
        -------
        X_new : np.ndarray
            The element matrices.
        """
        X_new = np.empty(shape=(n_series, self._window_size, n_windows))
        for i in range(n_series):
            rank = int(r[i])
            X_reduced = np.empty(shape=(rank, self._window_size, n_windows))
            for j in range(rank):
                X_reduced[j] = np.dot(np.outer(v[i, :, j], v[i, :, j]), X[i])
            X_new[i] = X_reduced.sum(axis=0)
        return X_new

    def _diagonal_averaging(
        self,
        X: np.ndarray,
        n_series: int,
        n_timestamps: int,
        n_windows: int,
        gap: int,
    ) -> np.ndarray:
        """
        Reconstruct the element matrices into time series.

        Parameters
        ----------
        X : np.ndarray
            The element matrices.
        n_metrics : int
            Number of time series (metrics).
        n_timestamps : int
            The length of time variable in the input data.
        n_windows : int
            The number of rows of the trajectory matrix.
        gap : int
            Either n_windows OR window_size.

        Returns
        -------
        X_new : np.ndarray
            The new time series (reflect signal only, no noise).
        """
        X_new = np.empty(shape=(n_series, n_timestamps))
        first_row = [(0, col) for col in range(n_windows)]
        last_col = [(row, n_windows - 1) for row in range(1, self._window_size)]
        indices = first_row + last_col
        for i in range(n_series):
            for j, k in indices:
                X_new[i, j + k] = np.diag(X[i, :, ::-1], gap - j - k - 1).mean()
        return X_new



from math import ceil
from typing import Optional, Union

import numpy as np
import pandas as pd
from scipy import stats
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array

from ._hpf import HighPassFilter
from ._ssa import SingularSpectrumAnalysis
from .base import BaseDetector


class Gripa(BaseEstimator, BaseDetector):
    """ """

    def __init__(
        self, window_size: Union[int, float] = 11, algorithm: str = 'hpf', **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self._window_size: Union[int, float] = window_size
        self._window_size_int: Optional[int] = None
        self._algorithm: str = algorithm

    def fit(self, data: Union[pd.DataFrame, pd.Series]) -> "BaseDetector":
        """ """
        self.validate_input_type(data)
        if data.isnull().values.any():  # type: ignore
            raise ValueError("Data contains missing values. Please handle it properly.")
        else:
            pass
        if isinstance(data, pd.Series):
            df = data.copy().to_frame()
        elif isinstance(data, pd.DataFrame):
            df = data.copy()
        n_timestamps = df.shape[0]
        self._window_size_int = self.validate_params(
            self._algorithm, self._window_size, n_timestamps
        )
        X = df.to_numpy().T
        X = check_array(X, dtype='float64')
        if self._algorithm == 'hpf':
            X_noise = HighPassFilter(self._window_size_int).transform(X)
        elif self._algorithm == 'ssa':
            X_noise = SingularSpectrumAnalysis(self._window_size_int).transform(X)
        self.anomaly_score = np.abs(stats.zscore(X_noise))
        if self.anomaly_score is not None:
            if isinstance(data, pd.DataFrame):
                self.anomaly_score = pd.DataFrame(
                    self.anomaly_score, index=data.index, columns=data.columns
                )
            elif isinstance(data, pd.Series):
                self.anomaly_score = pd.Series(
                    self.anomaly_score.ravel(),  # type: ignore
                    index=data.index,
                    name=data.name,
                )
        else:
            raise ValueError('The algorithm failed to produce anomaly scores.')
        return self

    @staticmethod
    def validate_input_type(data: Union[pd.DataFrame, pd.Series]) -> None:
        """ """
        if not isinstance(data, (pd.DataFrame, pd.Series)):
            raise TypeError('Input data must be pandas Series or DataFrame.')
        else:
            pass

    @staticmethod
    def validate_params(
        algorithm: str, window_size: Union[int, float], n_timestamps: int
    ) -> int:
        """ """
        if algorithm not in ['hpf', 'ssa']:
            raise ValueError('Parameter "algorithm" should be "hpf" or "ssa".')
        if not isinstance(window_size, (int, np.integer, float, np.floating)):
            raise TypeError('Parameter "window_size" should be integer or float.')
        if isinstance(window_size, (int, np.integer)):
            if not 2 <= window_size <= n_timestamps:
                raise ValueError(
                    'If "window_size" is an integer, it must be greater '
                    'than or equal to 2 and lower than or equal to '
                    'n_timestamps (got {}).'.format(window_size)
                )
            else:
                final_window_size = window_size
        else:
            if not 0 < window_size <= 1:
                raise ValueError(
                    'If "window_size" is a float, it must be greater '
                    'than 0 and lower than or equal to 1 '
                    '(got {}).'.format(window_size)
                )
            else:
                final_window_size = int(max(2, ceil(window_size * n_timestamps)))
        return final_window_size



import random

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal, assert_series_equal

from ..detector import Gripa

random.seed(10)


@pytest.fixture(scope="module")
def anomalous_frame_in():
    N = 40
    data1 = np.random.rand(N)
    data1[10] *= 10e3
    data2 = np.random.rand(N)
    data2[20] *= 10e3
    data3 = np.random.rand(N)
    data3[30] *= 10e3
    t = pd.date_range('2022-01-01', periods=N, freq='15T')
    frame = pd.DataFrame(
        {"metric1": data1, "metric2": data2, "metric3": data3}, index=t
    )
    return frame


@pytest.fixture(scope="module")
def anomalous_frame_out():
    N = 40
    data1 = [False] * N
    data1[10] = True
    data2 = [False] * N
    data2[20] = True
    data3 = [False] * N
    data3[30] = True
    t = pd.date_range('2022-01-01', periods=N, freq='15T')
    frame = pd.DataFrame(
        {"metric1": data1, "metric2": data2, "metric3": data3}, index=t
    )
    return frame


@pytest.fixture(scope="module")
def normal_frame_in():
    N = 40
    data1 = np.random.rand(N)
    data2 = np.random.rand(N)
    data3 = np.random.rand(N)
    t = pd.date_range('2022-01-01', periods=N, freq='15T')
    frame = pd.DataFrame(
        {"metric1": data1, "metric2": data2, "metric3": data3}, index=t
    )
    return frame


@pytest.fixture(scope="module")
def normal_frame_out():
    N = 40
    data1 = [False] * N
    data2 = [False] * N
    data3 = [False] * N
    t = pd.date_range('2022-01-01', periods=N, freq='15T')
    frame = pd.DataFrame(
        {"metric1": data1, "metric2": data2, "metric3": data3}, index=t
    )
    return frame


@pytest.fixture(scope="module")
def anomalous_series_in():
    N = 40
    data = np.random.rand(N)
    data[20] *= 10e3
    t = pd.date_range('2022-01-01', periods=N, freq='15T')
    series = pd.Series(data, name="metric1", index=t)
    return series


@pytest.fixture(scope="module")
def anomalous_series_out():
    N = 40
    data = [False] * N
    data[20] = True
    t = pd.date_range('2022-01-01', periods=N, freq='15T')
    series = pd.Series(data, name="metric1", index=t)
    return series


@pytest.fixture(scope="module")
def normal_series_in():
    N = 40
    data = np.random.rand(N)
    t = pd.date_range('2022-01-01', periods=N, freq='15T')
    series = pd.Series(data, name="metric1", index=t)
    return series


@pytest.fixture(scope="module")
def normal_series_out():
    N = 40
    data = [False] * N
    t = pd.date_range('2022-01-01', periods=N, freq='15T')
    series = pd.Series(data, name="metric1", index=t)
    return series


@pytest.fixture(scope="module")
def array_in():
    data = np.random.rand(40, 3)
    return data


@pytest.fixture(scope="module")
def missing_frame_in():
    N = 40
    data1 = np.random.rand(N)
    data1[10] = None
    data2 = np.random.rand(N)
    data2[20] = None
    data3 = np.random.rand(N)
    data3[30] = None
    t = pd.date_range('2022-01-01', periods=N, freq='15T')
    frame = pd.DataFrame(
        {"metric1": data1, "metric2": data2, "metric3": data3}, index=t
    )
    return frame


def test_anomalous_frame(
    anomalous_frame_in: pd.DataFrame, anomalous_frame_out: pd.DataFrame
) -> None:
    tsad_model = Gripa(algorithm='hpf')
    tsad_model.fit(anomalous_frame_in.copy())
    assert tsad_model.anomaly_score is not None
    assert tsad_model.anomaly_score.shape == anomalous_frame_in.shape
    assert all(tsad_model.anomaly_score.index == anomalous_frame_in.index)
    assert all(tsad_model.anomaly_score.columns == anomalous_frame_in.columns)
    anomalies = tsad_model.fit_predict(anomalous_frame_in.copy())
    assert anomalies is not None
    assert anomalies.shape == anomalous_frame_in.shape
    assert all(anomalies.index == anomalous_frame_in.index)
    assert all(anomalies.columns == anomalous_frame_in.columns)
    assert_frame_equal(
        anomalies,  # type: ignore
        anomalous_frame_out.copy(),
    )
    tsad_model = Gripa(algorithm='ssa')
    tsad_model.fit(anomalous_frame_in.copy())
    assert tsad_model.anomaly_score is not None
    assert tsad_model.anomaly_score.shape == anomalous_frame_in.shape
    assert all(tsad_model.anomaly_score.index == anomalous_frame_in.index)
    assert all(tsad_model.anomaly_score.columns == anomalous_frame_in.columns)
    anomalies = tsad_model.fit_predict(anomalous_frame_in.copy())
    assert anomalies is not None
    assert anomalies.shape == anomalous_frame_in.shape
    assert all(anomalies.index == anomalous_frame_in.index)
    assert all(anomalies.columns == anomalous_frame_in.columns)
    assert_frame_equal(
        anomalies,  # type: ignore
        anomalous_frame_out.copy(),
    )


def test_normal_frame(
    normal_frame_in: pd.DataFrame, normal_frame_out: pd.DataFrame
) -> None:
    tsad_model = Gripa(algorithm='hpf')
    tsad_model.fit(normal_frame_in.copy())
    assert tsad_model.anomaly_score is not None
    assert tsad_model.anomaly_score.shape == normal_frame_in.shape
    assert all(tsad_model.anomaly_score.index == normal_frame_in.index)
    assert all(tsad_model.anomaly_score.columns == normal_frame_in.columns)
    anomalies = tsad_model.fit_predict(normal_frame_in.copy())
    assert anomalies is not None
    assert anomalies.shape == normal_frame_in.shape
    assert all(anomalies.index == normal_frame_in.index)
    assert all(anomalies.columns == normal_frame_in.columns)
    assert_frame_equal(
        anomalies,  # type: ignore
        normal_frame_out.copy(),
    )
    tsad_model = Gripa(algorithm='ssa')
    tsad_model.fit(normal_frame_in.copy())
    assert tsad_model.anomaly_score is not None
    assert tsad_model.anomaly_score.shape == normal_frame_in.shape
    assert all(tsad_model.anomaly_score.index == normal_frame_in.index)
    assert all(tsad_model.anomaly_score.columns == normal_frame_in.columns)
    anomalies = tsad_model.fit_predict(normal_frame_in.copy())
    assert anomalies is not None
    assert anomalies.shape == normal_frame_in.shape
    assert all(anomalies.index == normal_frame_in.index)
    assert all(anomalies.columns == normal_frame_in.columns)
    assert_frame_equal(
        anomalies,  # type: ignore
        normal_frame_out.copy(),
    )


def test_anomalous_series(
    anomalous_series_in: pd.Series, anomalous_series_out: pd.Series
) -> None:
    tsad_model = Gripa(algorithm='hpf')
    tsad_model.fit(anomalous_series_in.copy())
    assert tsad_model.anomaly_score is not None
    assert tsad_model.anomaly_score.shape == anomalous_series_in.shape
    assert all(tsad_model.anomaly_score.index == anomalous_series_in.index)
    assert tsad_model.anomaly_score.name == anomalous_series_in.name
    anomalies = tsad_model.fit_predict(anomalous_series_in.copy())
    assert anomalies is not None
    assert anomalies.shape == anomalous_series_in.shape
    assert all(anomalies.index == anomalous_series_in.index)
    assert anomalies.name == anomalous_series_in.name
    assert_series_equal(
        anomalies,  # type: ignore
        anomalous_series_out.copy(),
    )
    tsad_model = Gripa(algorithm='ssa')
    tsad_model.fit(anomalous_series_in.copy())
    assert tsad_model.anomaly_score is not None
    assert tsad_model.anomaly_score.shape == anomalous_series_in.shape
    assert all(tsad_model.anomaly_score.index == anomalous_series_in.index)
    assert tsad_model.anomaly_score.name == anomalous_series_in.name
    anomalies = tsad_model.fit_predict(anomalous_series_in.copy())
    assert anomalies is not None
    assert anomalies.shape == anomalous_series_in.shape
    assert all(anomalies.index == anomalous_series_in.index)
    assert anomalies.name == anomalous_series_in.name
    assert_series_equal(
        anomalies,  # type: ignore
        anomalous_series_out.copy(),
    )


def test_normal_series(
    normal_series_in: pd.Series, normal_series_out: pd.Series
) -> None:
    tsad_model = Gripa(algorithm='hpf')
    tsad_model.fit(normal_series_in.copy())
    assert tsad_model.anomaly_score is not None
    assert tsad_model.anomaly_score.shape == normal_series_in.shape
    assert all(tsad_model.anomaly_score.index == normal_series_in.index)
    assert tsad_model.anomaly_score.name == normal_series_in.name
    anomalies = tsad_model.fit_predict(normal_series_in.copy())
    assert anomalies is not None
    assert anomalies.shape == normal_series_in.shape
    assert all(anomalies.index == normal_series_in.index)
    assert anomalies.name == normal_series_in.name
    assert_series_equal(
        anomalies,  # type: ignore
        normal_series_out.copy(),
    )
    tsad_model = Gripa(algorithm='ssa')
    tsad_model.fit(normal_series_in.copy())
    assert tsad_model.anomaly_score is not None
    assert tsad_model.anomaly_score.shape == normal_series_in.shape
    assert all(tsad_model.anomaly_score.index == normal_series_in.index)
    assert tsad_model.anomaly_score.name == normal_series_in.name
    anomalies = tsad_model.fit_predict(normal_series_in.copy())
    assert anomalies is not None
    assert anomalies.shape == normal_series_in.shape
    assert all(anomalies.index == normal_series_in.index)
    assert anomalies.name == normal_series_in.name
    assert_series_equal(
        anomalies,  # type: ignore
        normal_series_out.copy(),
    )


def test_input_type(array_in: np.ndarray) -> None:
    tsad_model = Gripa(algorithm='hpf')
    with pytest.raises(
        TypeError, match='Input data must be pandas Series or DataFrame.'
    ):
        tsad_model.fit(array_in.copy())  # type: ignore
    tsad_model = Gripa(algorithm='ssa')
    with pytest.raises(
        TypeError, match='Input data must be pandas Series or DataFrame.'
    ):
        tsad_model.fit(array_in.copy())  # type: ignore


def test_input_missing(missing_frame_in: pd.DataFrame) -> None:
    tsad_model = Gripa(algorithm='hpf')
    with pytest.raises(
        ValueError, match='Data contains missing values. Please handle it properly.'
    ):
        _ = tsad_model.fit_predict(missing_frame_in.copy())
    tsad_model = Gripa(algorithm='ssa')
    with pytest.raises(
        ValueError, match='Data contains missing values. Please handle it properly.'
    ):
        _ = tsad_model.fit_predict(missing_frame_in.copy())


def test_wrong_parameter(normal_series_in: pd.Series) -> None:
    tsad_model = Gripa(algorithm='random')
    with pytest.raises(
        ValueError, match='Parameter "algorithm" should be "hpf" or "ssa".'
    ):
        tsad_model.fit(normal_series_in)
    tsad_model = Gripa(window_size='random')  # type: ignore
    with pytest.raises(
        TypeError, match='Parameter "window_size" should be integer or float.'
    ):
        tsad_model.fit(normal_series_in)



import numpy as np

from .utils import rolling_window


class HighPassFilter:
    """High-Pass (HP) Filter.

    A filtering method that separates the high-frequency noise
    or residuals from time series. Inter-quartile range (IQR) mean
    is utilised to get the low-frequency signals (smoothed time series).

    Parameters
    ----------
    window_size : int or float (default=13)
        Size of the sliding window. If float, it represents
        the percentage of the size of each time series and must be
        between 0 and 1. The window size will be computed as
        ``max(2, ceil(window_size * n_timestamps))``.

    References
    ----------
    .. [1] Lee Redner.
            "A note on the use of nonlinear filtering in computer graphics"
            IEEE Computer Graphics and Applications, 10(3), pp. 23-29.
    """

    def __init__(
        self,
        window_size: int,
    ) -> None:
        self._window_size = window_size

    def transform(self, X: np.ndarray) -> np.ndarray:
        """
        Transform data using HP filter to extract noise or residuals.

        Parameters
        ----------
        data : pandas Series or DataFrame, shape = (n_timestamps, n_metrics)
            Input data.

        Returns
        -------
        X_noise : np.ndarray, shape = (n_timestamp, n_metrics)
            Extracted noise.
        """
        # 1. Constructing trajectory matrix
        n_series, n_timestamps = X.shape
        X_middle = rolling_window(X, n_series, n_timestamps, self._window_size).copy()
        # 2. Filling the starting & ending boundary
        n_top = int(np.ceil((self._window_size - 1) / 2))
        X_top = self._fix_starting_boundary(X, n_top)
        n_bottom = int(self._window_size - n_top - 1)
        X_bottom = self._fix_ending_boundary(X, n_bottom)
        # 3. Compiling the final signal
        X_window = np.concatenate((X_top, X_middle, X_bottom), axis=1)
        # 4. Calculating IQR mean
        median_id = self._window_size // 2
        q1_id = median_id // 2
        q3_id = median_id + (self._window_size - median_id) // 2
        X_window = np.sort(X_window, axis=2)
        X_signal = X_window[:, :, q1_id : q3_id + 1].mean(axis=2)
        X_noise = X - X_signal
        return X_noise.T

    def _fix_starting_boundary(
        self,
        X: np.ndarray,
        n_top: int,
    ) -> np.ndarray:
        """
        Fill the gap in the beginning of time series.

        Parameters
        ----------
        data : pandas Series or DataFrame, shape = (n_timestamps, n_series)
            Input data.
        n_top : int
            Number of gap in the starting time series.

        Returns
        -------
        X : np.ndarray
            Rolling-window matrices to fill the gap in the starting points.
        """
        n_series, _ = X.shape
        indices = np.array(range(self._window_size))
        n = self._window_size + n_top - 1
        extension = np.zeros(shape=(n_series, n), dtype=float)
        extension[:, n_top:] = X[:, indices[: self._window_size - 1]]
        extension[:, 0:n_top] = X[:, indices[n_top:0:-1]]
        X_top = rolling_window(extension, n_series, n, self._window_size).copy()
        return X_top

    def _fix_ending_boundary(
        self,
        X: np.ndarray,
        n_bottom: int,
    ) -> np.ndarray:
        """
        #TODO: change from pandas to numpy
        Fill the gap in the ending of time series.

        Parameters
        ----------
        data : pandas Series or DataFrame, shape = (n_timestamps, n_series)
            Input data.
        n_top : int
            Number of gap in the ending time series.

        Returns
        -------
        X : np.ndarray
            Rolling-window matrices to fill the gap in the ending points.
        """
        n_series, n_timestamps = X.shape
        indices = np.array(range(n_timestamps - self._window_size, n_timestamps))
        n = self._window_size + n_bottom - 1
        extension = np.zeros((n_series, n), float)
        extension[:, :-n_bottom] = X[:, indices[-(self._window_size - 1) :]]
        extension[:, -n_bottom:] = X[:, indices[-2 : -2 - n_bottom : -1]]
        X_bottom = rolling_window(extension, n_series, n, self._window_size).copy()
        return X_bottom
